module;

#include <cstdint>
#include <utility>

export module tile;

export enum class Direction : std::uint8_t {
	East = 0b00,
	North = 0b01,
	West = 0b10,
	South = 0b11,
};

export Direction
opposite(Direction direction) {
	switch (direction) {
	case Direction::East: return Direction::West;
	case Direction::North: return Direction::South;
	case Direction::West: return Direction::East;
	case Direction::South: return Direction::North;
	default: std::unreachable();
	}
}

export class Tile {
  public:
	static Tile const Blank;
	static Tile const Cross;
	static Tile const Slash;
	static Tile const Backslash;

	Direction connection(Direction entry) const;
	void remove_connection(Direction entry);
	bool empty() const { return good != 0b00; }

  private:
	std::underlying_type_t<Direction> direction : 2;
	std::uint8_t good : 2;

	Tile(Direction east) : direction{std::to_underlying(east)}, good{0b11} {}
};

Tile const Tile::Blank{Direction::East};
Tile const Tile::Cross{Direction::West};
Tile const Tile::Slash{Direction::South};
Tile const Tile::Backslash{Direction::North};

Direction
Tile::connection(Direction entry) const {
	Direction const exit = static_cast<Direction>(direction ^ std::to_underlying(entry));
	bool const is_good = (good >> (entry == Direction::East || exit == Direction::East)) & 1;
	return is_good ? exit : entry;
}

void
Tile::remove_connection(Direction entry) {
	Direction const exit = static_cast<Direction>(direction ^ std::to_underlying(entry));
	good &= ~(1 << (entry == Direction::East || exit == Direction::East));
}